#include "apt-cache-parser.h"

/* States of the parser. */
typedef enum {
  /* We are looking for a 'Package: ' header. */
  APT_CACHE_PARSER_STATE_WANT_PACKAGE,
  /* We are looking for the 'Section: ' header, to see if we have found a webapp */
  APT_CACHE_PARSER_STATE_WANT_SECTION,
  /* Once we have found a webapp we want to find 'Description: ' where the webapp metadata will be
     embedded */
  APT_CACHE_PARSER_STATE_WANT_DESCRIPTION,
  /* Once we have found the description field for a webapp we look for the metadata tag */
  APT_CACHE_PARSER_STATE_WANT_DESCRIPTION_METADATA,
  /* First contained in the metadata field is ' Name: <<APPLICATION_NAME>>' */
  APT_CACHE_PARSER_STATE_WANT_APPLICATION_NAME,
  /* In the metadata tag is a list of URL globs*/
  APT_CACHE_PARSER_STATE_WANT_URL_REGEX,
  APT_CACHE_PARSER_STATE_INVALID 
} AptCacheParserState;

struct _AptCacheParser {
  /* Current state of the parser */
  AptCacheParserState state;
  
  /* Package we are currently looking at. */
  gchar *last_package_name;
  gchar *last_application_name;
  
  /* Webapps we have found and parsed so far */
  GList *webapp_infos;
  
  /* Webapp we are currently building */
  WebappInfo *current_webapp;
};

AptCacheParser *
apt_cache_parser_new ()
{
  AptCacheParser *parser;
  
  parser = g_malloc0 (sizeof (AptCacheParser));
  parser->state = APT_CACHE_PARSER_STATE_WANT_PACKAGE;
  
  parser->last_package_name = NULL;
  
  parser->webapp_infos = NULL;
  
  parser->current_webapp = NULL;
  
  return parser;
}

static void
apt_cache_parser_free_webapp_infos (AptCacheParser *parser)
{
  GList *w;
  
  for (w = parser->webapp_infos; w != NULL; w = w->next)
    {
      WebappInfo *info;
      
      info = (WebappInfo *)w->data;
      webapp_info_free (info);
    }
  g_list_free (parser->webapp_infos);
  
  parser->webapp_infos = NULL;
}

void
apt_cache_parser_free (AptCacheParser *parser)
{
  apt_cache_parser_free_webapp_infos (parser);
  if (parser->current_webapp != NULL)
    {
      webapp_info_free (parser->current_webapp);
    }
  if (parser->last_package_name != NULL)
    {
      g_free (parser->last_package_name);
    }
  g_free (parser);
}

static gboolean
apt_cache_parser_parse_package_name (const gchar *line, gchar **out_name)
{
  const gchar *w;
  
  *out_name = NULL;
  // Just make this a precondition? kind of wasteful paranoia
  if (g_str_has_prefix (line, "Package: ") == FALSE)
    {
      return FALSE;
    }
  
  w = line;
  while (*w != ':' && *w != '\0')
    {
      w++;
    }
  if (*w == '\0')
    {
      return FALSE;
    }
  *out_name = g_strstrip (g_strdup(w+2)); // "Skip ': '"
  
  
  return TRUE;
}

static gboolean
apt_cache_parser_parse_application_name (const gchar *line, gchar **out_name)
{
  const gchar *w;
  
  *out_name = NULL;
  if (g_str_has_prefix (line, " Name: ") == FALSE)
    {
      return FALSE;
    }
  
  w = line;
  while (*w != ':' && *w != '\0')
    {
      w++;
    }
  if (*w == '\0')
    {
      return FALSE;
    }
  *out_name = g_strstrip (g_strdup(w+2)); // "Skip ': '"
  
  
  return TRUE;
}

static gboolean
apt_cache_parser_section_is_webapp (AptCacheParser *parser, const gchar *section_line)
{
  const gchar *w;
  gchar *section;
  gboolean ret;
  
  w = section_line;
  while (*w != ' ' && *w != '\0')
    {
      w++;
    }
  
  section = g_strdup(w);
  section = g_strstrip(section);
  
  ret = FALSE;

  if (g_strcmp0(section, "webapps") == 0)
    {
      ret = TRUE;
    }
  g_free (section);
  return ret;
}


gboolean
apt_cache_parser_next_line (AptCacheParser *parser, const gchar *line)
{
  switch (parser->state)
    {
    case APT_CACHE_PARSER_STATE_WANT_PACKAGE:
      if (g_str_has_prefix(line, "Package:"))
	{
	  gboolean parsed_name_field;
	  
	  if (parser->last_package_name != NULL)
	    {
	      g_free (parser->last_package_name);
	      parser->last_package_name = NULL;
	    }
	  parsed_name_field = apt_cache_parser_parse_package_name (line, &parser->last_package_name);
	  
	  if (parsed_name_field == FALSE)
	    {
	      g_error ("Failed to parse name field of package index");
	    }

	  parser->state = APT_CACHE_PARSER_STATE_WANT_SECTION;
	  return TRUE;
	}
      break;
    case APT_CACHE_PARSER_STATE_WANT_SECTION:
      if (g_str_has_prefix(line, "Section:"))
	{
	  gboolean is_webapp;
	  
	  is_webapp = apt_cache_parser_section_is_webapp (parser, line);
	  
	  if (is_webapp == FALSE)
	    {
	      parser->state = APT_CACHE_PARSER_STATE_WANT_PACKAGE;
	      return TRUE;
	    }
	  else
	    {
	      parser->current_webapp = webapp_info_new (parser->last_package_name);
	      parser->state = APT_CACHE_PARSER_STATE_WANT_DESCRIPTION;
	      return TRUE;
	    }
	  return TRUE;
	}
      break;
    case APT_CACHE_PARSER_STATE_WANT_DESCRIPTION:
      //What is invalid here?
      if (g_str_has_prefix(line, "Description: ")) 
	{
	  parser->state = APT_CACHE_PARSER_STATE_WANT_DESCRIPTION_METADATA;
	}
      break;
    case APT_CACHE_PARSER_STATE_WANT_DESCRIPTION_METADATA:
      //What is invalid here?
      if (g_str_has_prefix(line, " <<ubuntu-webapps-metadata>>")) 
	{
	  parser->state = APT_CACHE_PARSER_STATE_WANT_APPLICATION_NAME;
	}
      break;
    case APT_CACHE_PARSER_STATE_WANT_APPLICATION_NAME:
      if (parser->last_application_name != NULL)
	{
	  g_free (parser->last_application_name);
	  parser->last_application_name = NULL;
	}
      gboolean parsed_name;

      parsed_name = apt_cache_parser_parse_application_name (line, &parser->last_application_name);
      if (parsed_name == FALSE)
	{
	  g_critical ("Failed to parse webapp application name metadata field: %s\n", parser->last_package_name);
	  g_critical ("Skipping application\n");
	  webapp_info_free (parser->current_webapp);
	  parser->current_webapp = NULL;

	  parser->state = APT_CACHE_PARSER_STATE_WANT_PACKAGE;
	  break;
	}
      webapp_info_set_application_name (parser->current_webapp, parser->last_application_name);
      parser->state = APT_CACHE_PARSER_STATE_WANT_URL_REGEX;
      break;
    case APT_CACHE_PARSER_STATE_WANT_URL_REGEX:
      if (g_str_has_prefix(line, " <<ubuntu-webapps-metadata>>"))
	{
	  parser->state = APT_CACHE_PARSER_STATE_WANT_PACKAGE;
	  parser->webapp_infos = g_list_append (parser->webapp_infos, parser->current_webapp);
	  parser->current_webapp = NULL;
	}
      else if (*line == ' ')
	{
	  webapp_info_add_url (parser->current_webapp, line+1);
	}
      break;
    case APT_CACHE_PARSER_STATE_INVALID:
      g_assert_not_reached();
    }
  return TRUE;
}
  
GList *
apt_cache_parser_get_webapp_infos (AptCacheParser *parser)
{
  return parser->webapp_infos;
}
